use std::marker::PhantomData;
use std::str::Bytes;

pub trait IsSigned {}

impl IsSigned for i8 {}
impl IsSigned for i16 {}
impl IsSigned for i32 {}
impl IsSigned for i64 {}
impl IsSigned for i128 {}
impl IsSigned for isize {}

pub trait IsUnsigned {}

impl IsUnsigned for u8 {}
impl IsUnsigned for u16 {}
impl IsUnsigned for u32 {}
impl IsUnsigned for u64 {}
impl IsUnsigned for u128 {}
impl IsUnsigned for usize {}

pub struct Signed<'a, T = i64>
where
    T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned,
{
    bytes: Bytes<'a>,
    _phantom: PhantomData<T>,
}

impl<'a, T> Iterator for Signed<'a, T>
where
    T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned,
{
    type Item = T;

    fn next(&mut self) -> Option<Self::Item> {
        try_parse_signed::<T>(&mut self.bytes)
    }
}

pub struct Unsigned<'a, T = u64>
where
    T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned,
{
    bytes: Bytes<'a>,
    _phantom: PhantomData<T>,
}

impl<'a, T> Iterator for Unsigned<'a, T>
where
    T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned,
{
    type Item = T;

    fn next(&mut self) -> Option<Self::Item> {
        try_parse_unsigned::<T>(&mut self.bytes)
    }
}

pub trait ParseExt {
    fn parse_signed<T>(&self) -> T
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned;

    fn parse_unsigned<T>(&self) -> T
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned;

    fn signed<T>(&self) -> Signed<'_, T>
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned;

    fn unsigned<T>(&self) -> Unsigned<'_, T>
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned;
}

impl ParseExt for &str {
    fn signed<T>(&self) -> Signed<'_, T>
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned,
    {
        Signed {
            bytes: self.bytes(),
            _phantom: PhantomData,
        }
    }

    fn parse_signed<T>(&self) -> T
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned,
    {
        match try_parse_signed(&mut self.bytes()) {
            Some(num) => num,
            None => panic!("unable to parse signed: {self}"),
        }
    }

    fn unsigned<T>(&self) -> Unsigned<'_, T>
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned,
    {
        Unsigned {
            bytes: self.bytes(),
            _phantom: PhantomData,
        }
    }

    fn parse_unsigned<T>(&self) -> T
    where
        T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned,
    {
        match try_parse_unsigned(&mut self.bytes()) {
            Some(num) => num,
            None => panic!("unable to parse signed: {self}"),
        }
    }
}

use std::ops::{Add, Mul};
fn try_parse_unsigned<T>(bytes: &mut Bytes<'_>) -> Option<T>
where
    T: Add<Output = T> + Mul<Output = T> + TryFrom<u8> + IsUnsigned,
{
    let mut n = loop {
        let byte = bytes.next()?;
        let digit = byte.wrapping_sub(b'0');

        if digit < 10 {
            break T::try_from(digit).unwrap_or_else(|_| unreachable!());
        }
    };

    loop {
        let Some(byte) = bytes.next() else {
            break Some(n);
        };
        let digit = byte.wrapping_sub(b'0');

        if digit < 10 {
            n = T::try_from(10).unwrap_or_else(|_| unreachable!()) * n
                + T::try_from(digit).unwrap_or_else(|_| unreachable!());
        } else {
            break Some(n);
        }
    }
}

fn try_parse_signed<T>(bytes: &mut Bytes<'_>) -> Option<T>
where
    T: Add<Output = T> + Mul<Output = T> + TryFrom<i32> + TryFrom<u8> + IsSigned,
{
    let mut fac = T::try_from(1i32).unwrap_or_else(|_| unreachable!());
    let mut n = loop {
        let byte = bytes.next()?;
        if byte == b'-' {
            let byte = bytes.next()?;
            let digit = byte.wrapping_sub(b'0');
            if digit < 10 {
                fac = T::try_from(-1i32).unwrap_or_else(|_| unreachable!());
                break T::try_from(digit).unwrap_or_else(|_| unreachable!());
            }
        }
        let digit = byte.wrapping_sub(b'0');

        if digit < 10 {
            break T::try_from(digit).unwrap_or_else(|_| unreachable!());
        }
    };

    loop {
        let Some(byte) = bytes.next() else {
            break Some(n * fac);
        };
        let digit = byte.wrapping_sub(b'0');

        if digit < 10 {
            n = T::try_from(10).unwrap_or_else(|_| unreachable!()) * n
                + T::try_from(digit).unwrap_or_else(|_| unreachable!());
        } else {
            break Some(n * fac);
        }
    }
}
